@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2020 gnome-mpv
+ * Copyright (c) 2014-2022 gnome-mpv
  *
  * This file is part of Celluloid.
  *
@@ -22,7 +22,6 @@
 #include <gio/gio.h>
 #include <gdk/gdk.h>
 
-#include "celluloid-main-window-private.h"
 #include "celluloid-def.h"
 #include "celluloid-marshal.h"
 #include "celluloid-menu.h"
@@ -33,6 +32,45 @@
 #include "celluloid-control-box.h"
 #include "celluloid-video-area.h"
 
+#define get_private(window) \
+	((CelluloidMainWindowPrivate *)celluloid_main_window_get_instance_private(CELLULOID_MAIN_WINDOW(window)))
+
+typedef struct _CelluloidMainWindowPrivate CelluloidMainWindowPrivate;
+
+enum
+{
+	PROP_0,
+	PROP_ALWAYS_FLOATING,
+	PROP_CHAPTER_LIST,
+	N_PROPERTIES
+};
+
+struct _CelluloidMainWindowPrivate
+{
+	GtkApplicationWindow parent;
+	gint width_offset;
+	gint height_offset;
+	gint resize_target[2];
+	gint compact_threshold;
+	gboolean csd;
+	gboolean always_floating;
+	gboolean use_floating_controls;
+	gboolean playlist_visible;
+	gboolean playlist_first_toggle;
+	gboolean pre_fs_playlist_visible;
+	gint playlist_width;
+	guint resize_tag;
+	GPtrArray *chapter_list;
+	const GPtrArray *track_list;
+	const GPtrArray *disc_list;
+	GtkWidget *header_bar;
+	GtkWidget *main_box;
+	GtkWidget *video_area_paned;
+	GtkWidget *video_area;
+	GtkWidget *control_box;
+	GtkWidget *playlist;
+};
+
 static void
 constructed(GObject *object);
 
@@ -51,6 +89,9 @@ get_property(	GObject *object,
 		GValue *value,
 		GParamSpec *pspec );
 
+static void
+notify_fullscreened_handler(GObject *object, GParamSpec *pspec, gpointer data);
+
 static void
 seek_handler(GtkWidget *widget, gdouble value, gpointer data);
 
@@ -62,14 +103,21 @@ button_clicked_handler(	CelluloidControlBox *control_box,
 static void
 notify(GObject *object, GParamSpec *pspec);
 
+static void
+size_allocate(GtkWidget *widget, gint width, gint height, gint baseline);
+
 static void
 resize_video_area_finalize(	GtkWidget *widget,
-				GdkRectangle *allocation,
+				gint width,
+				gint height,
 				gpointer data );
 
 static gboolean
 resize_to_target(gpointer data);
 
+gboolean
+update_compact_mode(GtkWidget *widget);
+
 G_DEFINE_TYPE_WITH_PRIVATE(CelluloidMainWindow, celluloid_main_window, GTK_TYPE_APPLICATION_WINDOW)
 
 static void
@@ -79,22 +127,20 @@ constructed(GObject *object)
 
 	priv->playlist = celluloid_playlist_widget_new();
 
-	gtk_widget_show_all(priv->playlist);
 	gtk_widget_hide(priv->playlist);
-	gtk_widget_set_no_show_all(priv->playlist, TRUE);
-
-	gtk_widget_show_all(priv->control_box);
 	gtk_widget_hide(priv->control_box);
-	gtk_widget_set_no_show_all(priv->control_box, TRUE);
 
-	gtk_paned_pack1(	GTK_PANED(priv->vid_area_paned),
-				priv->vid_area,
-				TRUE,
-				TRUE );
-	gtk_paned_pack2(	GTK_PANED(priv->vid_area_paned),
-				priv->playlist,
-				FALSE,
-				FALSE );
+	gtk_paned_set_start_child
+		(GTK_PANED(priv->video_area_paned), priv->video_area);
+	gtk_paned_set_end_child
+		(GTK_PANED(priv->video_area_paned), priv->playlist);
+	gtk_paned_set_shrink_start_child
+		(GTK_PANED(priv->video_area_paned), FALSE);
+	gtk_paned_set_shrink_end_child
+		(GTK_PANED(priv->video_area_paned), FALSE);
+
+	gtk_application_window_set_show_menubar
+		(GTK_APPLICATION_WINDOW(object), !priv->csd);
 
 	G_OBJECT_CLASS(celluloid_main_window_parent_class)->constructed(object);
 }
@@ -119,6 +165,10 @@ set_property(	GObject *object,
 	{
 		priv->always_floating = g_value_get_boolean(value);
 	}
+	else if(property_id == PROP_CHAPTER_LIST)
+	{
+		priv->chapter_list = g_value_get_pointer(value);
+	}
 	else
 	{
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
@@ -137,12 +187,62 @@ get_property(	GObject *object,
 	{
 		g_value_set_boolean(value, priv->always_floating);
 	}
+	else if(property_id == PROP_ALWAYS_FLOATING)
+	{
+		g_value_set_pointer(value, priv->chapter_list);
+	}
 	else
 	{
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
 	}
 }
 
+static void
+notify_fullscreened_handler(GObject *object, GParamSpec *pspec, gpointer data)
+{
+	CelluloidMainWindow *wnd =
+		CELLULOID_MAIN_WINDOW(object);
+	CelluloidMainWindowPrivate *priv =
+		get_private(wnd);
+	GSettings *settings =
+		g_settings_new(CONFIG_WIN_STATE);
+	const gboolean fullscreen =
+		gtk_window_is_fullscreen(GTK_WINDOW(object));
+	const gboolean floating =
+		priv->always_floating || fullscreen;
+	const gboolean playlist_visible =
+		!fullscreen && priv->pre_fs_playlist_visible;
+	const gboolean show_controls = g_settings_get_boolean
+		(settings, "show-controls");
+
+	if(fullscreen)
+	{
+		gtk_window_fullscreen(GTK_WINDOW(wnd));
+		gtk_window_present(GTK_WINDOW(wnd));
+
+		priv->pre_fs_playlist_visible = priv->playlist_visible;
+	}
+	else
+	{
+		gtk_window_unfullscreen(GTK_WINDOW(wnd));
+
+		priv->playlist_visible = priv->pre_fs_playlist_visible;
+	}
+
+	if(!celluloid_main_window_get_csd_enabled(wnd))
+	{
+		gtk_application_window_set_show_menubar
+			(GTK_APPLICATION_WINDOW(wnd), !fullscreen);
+	}
+
+	celluloid_main_window_set_use_floating_controls
+		(wnd, floating && show_controls);
+	gtk_widget_set_visible
+		(priv->playlist, playlist_visible);
+
+	g_object_unref(settings);
+}
+
 static void
 seek_handler(GtkWidget *widget, gdouble value, gpointer data)
 {
@@ -167,28 +267,38 @@ notify(GObject *object, GParamSpec *pspec)
 	{
 		CelluloidMainWindow *wnd = CELLULOID_MAIN_WINDOW(object);
 		CelluloidMainWindowPrivate *priv = get_private(wnd);
-		gboolean floating = priv->always_floating || priv->fullscreen;
+		gboolean floating =
+			priv->always_floating ||
+			gtk_window_is_fullscreen(GTK_WINDOW(object));
 
 		celluloid_main_window_set_use_floating_controls(wnd, floating);
 	}
 }
 
+static void
+size_allocate(GtkWidget *widget, gint width, gint height, gint baseline)
+{
+	GTK_WIDGET_CLASS(celluloid_main_window_parent_class)
+		->size_allocate(widget, width, height, baseline);
+
+	g_idle_add((GSourceFunc)update_compact_mode, widget);
+}
+
 static void
 resize_video_area_finalize(	GtkWidget *widget,
-				GdkRectangle *allocation,
+				gint width,
+				gint height,
 				gpointer data )
 {
 	CelluloidMainWindow *wnd = data;
 	CelluloidMainWindowPrivate *priv = get_private(wnd);
-	GdkWindow *gdk_window = gtk_widget_get_window(data);
+	GdkSurface *surface = gtk_widget_get_surface(data);
 	GdkDisplay *display = gdk_display_get_default();
-	GdkMonitor *monitor =	gdk_display_get_monitor_at_window
-				(display, gdk_window);
+	GdkMonitor *monitor = gdk_display_get_monitor_at_surface(display, surface);
 	GdkRectangle monitor_geom = {0};
-	gint width = allocation->width;
-	gint height = allocation->height;
 	gint target_width = priv->resize_target[0];
 	gint target_height = priv->resize_target[1];
+	const gint scale = gdk_surface_get_scale_factor(surface);
 
 	g_signal_handlers_disconnect_by_func
 		(widget, resize_video_area_finalize, data);
@@ -196,14 +306,14 @@ resize_video_area_finalize(	GtkWidget *widget,
 	gdk_monitor_get_geometry(monitor, &monitor_geom);
 
 	/* Adjust resize offset */
-	if((width != target_width || height != target_height)
+	if((width / scale != target_width || height / scale != target_height)
 	&& (	target_width < monitor_geom.width &&
 		target_height < monitor_geom.height )
 	&& !gtk_window_is_maximized(GTK_WINDOW(wnd))
-	&& !priv->fullscreen)
+	&& !gtk_window_is_fullscreen(GTK_WINDOW(wnd)))
 	{
-		priv->width_offset += target_width-width;
-		priv->height_offset += target_height-height;
+		priv->width_offset += target_width - width / scale;
+		priv->height_offset += target_height - height / scale;
 
 		g_source_clear(&priv->resize_tag);
 		priv->resize_tag = g_idle_add_full(	G_PRIORITY_HIGH_IDLE,
@@ -223,12 +333,10 @@ resize_to_target(gpointer data)
 
 	g_source_clear(&priv->resize_tag);
 
-	/* Emulate mpv resizing behavior by using GDK_GRAVITY_CENTER */
-	gtk_window_set_gravity(GTK_WINDOW(wnd), GDK_GRAVITY_CENTER);
-	gtk_window_resize(	GTK_WINDOW(wnd),
-				target_width+priv->width_offset,
-				target_height+priv->height_offset );
-	gtk_window_set_gravity(GTK_WINDOW(wnd), GDK_GRAVITY_NORTH_WEST);
+	gtk_window_set_default_size
+		(	GTK_WINDOW(wnd),
+			MAX(0, target_width + priv->width_offset),
+			MAX(0, target_height + priv->height_offset) );
 
 	/* Prevent graphical glitches that appear when calling
 	 * celluloid_main_window_resize_video_area() with the current size as
@@ -240,10 +348,39 @@ resize_to_target(gpointer data)
 	return FALSE;
 }
 
+gboolean
+update_compact_mode(GtkWidget *widget)
+{
+	CelluloidMainWindow *wnd = CELLULOID_MAIN_WINDOW(widget);
+	CelluloidMainWindowPrivate *priv = get_private(wnd);
+
+	if(priv->compact_threshold < 0)
+	{
+		gtk_widget_measure
+			(	widget,
+				GTK_ORIENTATION_HORIZONTAL,
+				-1,
+				&priv->compact_threshold,
+				NULL,
+				NULL,
+				NULL );
+
+		g_assert(priv->compact_threshold > -1);
+		priv->compact_threshold += COMPACT_THRESHOLD_OFFSET;
+	}
+
+	const gint width = gtk_widget_get_allocated_width(widget);
+	const gboolean compact = width <= priv->compact_threshold;
+	g_object_set(priv->control_box, "compact", compact, NULL);
+
+	return G_SOURCE_REMOVE;
+}
+
 static void
 celluloid_main_window_class_init(CelluloidMainWindowClass *klass)
 {
 	GObjectClass *obj_class = G_OBJECT_CLASS(klass);
+	GtkWidgetClass *wgt_class = GTK_WIDGET_CLASS(klass);
 	GParamSpec *pspec = NULL;
 
 	obj_class->constructed = constructed;
@@ -251,6 +388,7 @@ celluloid_main_window_class_init(CelluloidMainWindowClass *klass)
 	obj_class->set_property = set_property;
 	obj_class->get_property = get_property;
 	obj_class->notify = notify;
+	wgt_class->size_allocate = size_allocate;
 
 	pspec = g_param_spec_boolean
 		(	"always-use-floating-controls",
@@ -260,6 +398,13 @@ celluloid_main_window_class_init(CelluloidMainWindowClass *klass)
 			G_PARAM_READWRITE );
 	g_object_class_install_property(obj_class, PROP_ALWAYS_FLOATING, pspec);
 
+	pspec = g_param_spec_pointer
+		(	"chapter-list",
+			"Chapter list",
+			"The list of chapters for the current file",
+			G_PARAM_READWRITE );
+	g_object_class_install_property(obj_class, PROP_CHAPTER_LIST, pspec);
+
 	g_signal_new(	"button-clicked",
 			G_TYPE_FROM_CLASS(klass),
 			G_SIGNAL_RUN_FIRST|G_SIGNAL_DETAILED,
@@ -285,59 +430,89 @@ static void
 celluloid_main_window_init(CelluloidMainWindow *wnd)
 {
 	CelluloidMainWindowPrivate *priv = get_private(wnd);
-	CelluloidControlBox *vid_area_control_box = NULL;
+	CelluloidControlBox *video_area_control_box = NULL;
+	CelluloidHeaderBar *video_area_header_bar = NULL;
+	GSettings *state = g_settings_new(CONFIG_WIN_STATE);
+	GSettings *settings = g_settings_new(CONFIG_ROOT);
 
 	priv->csd = FALSE;
 	priv->always_floating = FALSE;
 	priv->use_floating_controls = FALSE;
-	priv->fullscreen = FALSE;
 	priv->playlist_visible = FALSE;
 	priv->pre_fs_playlist_visible = FALSE;
 	priv->playlist_width = PLAYLIST_DEFAULT_WIDTH;
 	priv->resize_tag = 0;
+	priv->chapter_list = NULL;
 	priv->track_list = NULL;
 	priv->disc_list = NULL;
 	priv->header_bar = celluloid_header_bar_new();
 	priv->main_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
-	priv->vid_area_paned = gtk_paned_new(GTK_ORIENTATION_HORIZONTAL);
-	priv->vid_area = celluloid_video_area_new();
+	priv->video_area_paned = gtk_paned_new(GTK_ORIENTATION_HORIZONTAL);
+	priv->video_area = celluloid_video_area_new();
 	priv->control_box = celluloid_control_box_new();
 
 	priv->playlist_first_toggle = TRUE;
 	priv->width_offset = 0;
 	priv->height_offset = 0;
+	priv->compact_threshold = -1;
 
-	vid_area_control_box =	celluloid_video_area_get_control_box
-				(CELLULOID_VIDEO_AREA(priv->vid_area));
+	video_area_control_box =
+		celluloid_video_area_get_control_box
+		(CELLULOID_VIDEO_AREA(priv->video_area));
+	video_area_header_bar =
+		celluloid_video_area_get_header_bar
+		(CELLULOID_VIDEO_AREA(priv->video_area));
 
-	g_object_bind_property(	wnd, "title",
-				priv->vid_area, "title",
+	g_settings_bind(	state, "loop-playlist",
+				priv->control_box, "loop",
+				G_SETTINGS_BIND_DEFAULT );
+	g_settings_bind(	settings, "menubar-accel-enable",
+				wnd, "handle-menubar-accel",
+				G_SETTINGS_BIND_DEFAULT );
+
+	g_object_bind_property(	wnd, "fullscreened",
+				priv->video_area, "fullscreened",
+				G_BINDING_DEFAULT );
+
+	g_object_bind_property(	priv->header_bar, "open-button-active",
+				video_area_header_bar, "open-button-active",
+				G_BINDING_DEFAULT );
+	g_object_bind_property(	priv->header_bar, "menu-button-active",
+				video_area_header_bar, "menu-button-active",
+				G_BINDING_DEFAULT );
+
+	g_object_bind_property(	priv->control_box, "chapter-list",
+				video_area_control_box, "chapter-list",
 				G_BINDING_DEFAULT );
 	g_object_bind_property(	priv->control_box, "duration",
-				vid_area_control_box, "duration",
+				video_area_control_box, "duration",
 				G_BINDING_DEFAULT );
 	g_object_bind_property(	priv->control_box, "pause",
-				vid_area_control_box, "pause",
+				video_area_control_box, "pause",
 				G_BINDING_DEFAULT );
 	g_object_bind_property(	priv->control_box, "skip-enabled",
-				vid_area_control_box, "skip-enabled",
+				video_area_control_box, "skip-enabled",
 				G_BINDING_DEFAULT );
 	g_object_bind_property(	priv->control_box, "show-fullscreen-button",
-				vid_area_control_box, "show-fullscreen-button",
+				video_area_control_box, "show-fullscreen-button",
 				G_BINDING_DEFAULT );
 	g_object_bind_property(	priv->control_box, "time-position",
-				vid_area_control_box, "time-position",
+				video_area_control_box, "time-position",
 				G_BINDING_DEFAULT );
 	g_object_bind_property(	priv->control_box, "loop",
-				vid_area_control_box, "loop",
+				video_area_control_box, "loop",
 				G_BINDING_BIDIRECTIONAL );
 	g_object_bind_property(	priv->control_box, "volume",
-				vid_area_control_box, "volume",
+				video_area_control_box, "volume",
 				G_BINDING_BIDIRECTIONAL );
 	g_object_bind_property(	priv->control_box, "volume-popup-visible",
-				vid_area_control_box, "volume-popup-visible",
+				video_area_control_box, "volume-popup-visible",
 				G_BINDING_BIDIRECTIONAL );
 
+	g_signal_connect(	wnd,
+				"notify::fullscreened",
+				G_CALLBACK(notify_fullscreened_handler),
+				wnd );
 	g_signal_connect(	priv->control_box,
 				"seek",
 				G_CALLBACK(seek_handler),
@@ -346,33 +521,33 @@ celluloid_main_window_init(CelluloidMainWindow *wnd)
 				"button-clicked",
 				G_CALLBACK(button_clicked_handler),
 				wnd );
-	g_signal_connect(	vid_area_control_box,
+	g_signal_connect(	video_area_control_box,
 				"seek",
 				G_CALLBACK(seek_handler),
 				wnd );
-	g_signal_connect(	vid_area_control_box,
+	g_signal_connect(	video_area_control_box,
 				"button-clicked",
 				G_CALLBACK(button_clicked_handler),
 				wnd );
 
-	gtk_widget_add_events(	priv->vid_area,
-				GDK_ENTER_NOTIFY_MASK
-				|GDK_LEAVE_NOTIFY_MASK );
-
 	gtk_window_set_title(GTK_WINDOW(wnd), g_get_application_name());
 
-	gtk_paned_set_position(	GTK_PANED(priv->vid_area_paned),
+	gtk_paned_set_position(	GTK_PANED(priv->video_area_paned),
 				MAIN_WINDOW_DEFAULT_WIDTH
 				-PLAYLIST_DEFAULT_WIDTH );
+	gtk_paned_set_resize_end_child(GTK_PANED(priv->video_area_paned), FALSE);
 
 	gtk_window_set_default_size(	GTK_WINDOW(wnd),
 					MAIN_WINDOW_DEFAULT_WIDTH,
 					MAIN_WINDOW_DEFAULT_HEIGHT );
 
-	gtk_box_pack_start
-		(GTK_BOX(priv->main_box), priv->vid_area_paned, TRUE, TRUE, 0);
-	gtk_container_add(GTK_CONTAINER(priv->main_box), priv->control_box);
-	gtk_container_add(GTK_CONTAINER(wnd), priv->main_box);
+	gtk_widget_set_hexpand(priv->header_bar, TRUE);
+	gtk_widget_set_hexpand(priv->control_box, TRUE);
+	gtk_widget_set_vexpand(priv->video_area_paned, TRUE);
+
+	gtk_box_append(GTK_BOX(priv->main_box), priv->video_area_paned);
+	gtk_box_append(GTK_BOX(priv->main_box), priv->control_box);
+	gtk_window_set_child(GTK_WINDOW(wnd), priv->main_box);
 }
 
 GtkWidget *
@@ -402,7 +577,7 @@ celluloid_main_window_get_control_box(CelluloidMainWindow *wnd)
 CelluloidVideoArea *
 celluloid_main_window_get_video_area(CelluloidMainWindow *wnd)
 {
-	return CELLULOID_VIDEO_AREA(get_private(wnd)->vid_area);
+	return CELLULOID_VIDEO_AREA(get_private(wnd)->video_area);
 }
 
 void
@@ -420,7 +595,7 @@ celluloid_main_window_set_use_floating_controls(	CelluloidMainWindow *wnd,
 		gtk_widget_set_visible
 			(priv->control_box, controls_visible && !floating);
 		celluloid_video_area_set_control_box_visible
-			(CELLULOID_VIDEO_AREA(priv->vid_area), floating);
+			(CELLULOID_VIDEO_AREA(priv->video_area), floating);
 
 		priv->use_floating_controls = floating;
 
@@ -434,58 +609,18 @@ celluloid_main_window_get_use_floating_controls(CelluloidMainWindow *wnd)
 	return get_private(wnd)->use_floating_controls;
 }
 
-void
-celluloid_main_window_set_fullscreen(CelluloidMainWindow *wnd, gboolean fullscreen)
-{
-	CelluloidMainWindowPrivate *priv = get_private(wnd);
-
-	if(fullscreen != priv->fullscreen)
-	{
-		CelluloidVideoArea *vid_area =
-			CELLULOID_VIDEO_AREA(priv->vid_area);
-		gboolean floating =
-			priv->always_floating || fullscreen;
-		gboolean playlist_visible =
-			!fullscreen && priv->pre_fs_playlist_visible;
-
-		if(fullscreen)
-		{
-			gtk_window_fullscreen(GTK_WINDOW(wnd));
-			gtk_window_present(GTK_WINDOW(wnd));
-
-			priv->pre_fs_playlist_visible = priv->playlist_visible;
-		}
-		else
-		{
-			gtk_window_unfullscreen(GTK_WINDOW(wnd));
-
-			priv->playlist_visible = priv->pre_fs_playlist_visible;
-		}
-
-		if(!celluloid_main_window_get_csd_enabled(wnd))
-		{
-			gtk_application_window_set_show_menubar
-				(GTK_APPLICATION_WINDOW(wnd), !fullscreen);
-		}
-
-		celluloid_video_area_set_fullscreen_state(vid_area, fullscreen);
-		celluloid_main_window_set_use_floating_controls(wnd, floating);
-		gtk_widget_set_visible(priv->playlist, playlist_visible);
-
-		priv->fullscreen = fullscreen;
-	}
-}
-
 gboolean
 celluloid_main_window_get_fullscreen(CelluloidMainWindow *wnd)
 {
-	return get_private(wnd)->fullscreen;
+	return gtk_window_is_fullscreen(GTK_WINDOW(wnd));
 }
 
 void
 celluloid_main_window_toggle_fullscreen(CelluloidMainWindow *wnd)
 {
-	celluloid_main_window_set_fullscreen(wnd, !get_private(wnd)->fullscreen);
+	const gboolean fullscreen = gtk_window_is_fullscreen(GTK_WINDOW(wnd));
+
+	g_object_set(wnd, "fullscreened", !fullscreen, NULL);
 }
 
 void
@@ -507,22 +642,20 @@ celluloid_main_window_save_state(CelluloidMainWindow *wnd)
 	gboolean maximized;
 	gint handle_pos;
 	gdouble volume;
-	gboolean loop;
-	gboolean controls_visible;
 
 	settings = g_settings_new(CONFIG_WIN_STATE);
 	priv = get_private(wnd);
 	maximized = gtk_window_is_maximized(GTK_WINDOW(wnd));
-	handle_pos = gtk_paned_get_position(GTK_PANED(priv->vid_area_paned));
-	controls_visible = gtk_widget_get_visible(priv->control_box);
+	handle_pos = gtk_paned_get_position(GTK_PANED(priv->video_area_paned));
 
-	g_object_get(priv->control_box, "volume", &volume, "loop", &loop, NULL);
-	gtk_window_get_size(GTK_WINDOW(wnd), &width, &height);
+	g_object_get(priv->control_box, "volume", &volume, NULL);
+	gtk_window_get_default_size(GTK_WINDOW(wnd), &width, &height);
 
+	// Controls visibility does not need to be saved here since
+	// celluloid_main_window_set_controls_visible() already updates the
+	// associated GSettings key when it is called.
 	g_settings_set_boolean(settings, "maximized", maximized);
 	g_settings_set_double(settings, "volume", volume/100.0);
-	g_settings_set_boolean(settings, "loop-playlist", loop);
-	g_settings_set_boolean(settings, "show-controls", controls_visible);
 	g_settings_set_boolean(settings, "show-playlist", priv->playlist_visible);
 
 	if(!maximized)
@@ -560,7 +693,6 @@ celluloid_main_window_load_state(CelluloidMainWindow *wnd)
 		gint handle_pos;
 		gboolean controls_visible;
 		gdouble volume;
-		gboolean loop;
 
 		priv->playlist_width
 			= g_settings_get_int(settings, "playlist-width");
@@ -569,19 +701,15 @@ celluloid_main_window_load_state(CelluloidMainWindow *wnd)
 		controls_visible
 			= g_settings_get_boolean(settings, "show-controls");
 		volume = g_settings_get_double(settings, "volume");
-		loop = g_settings_get_boolean(settings, "loop-playlist");
 		handle_pos = width-(priv->playlist_visible?priv->playlist_width:0);
 
-		g_object_set(	priv->control_box,
-				"volume", volume,
-				"loop", loop,
-				NULL);
+		g_object_set(priv->control_box, "volume", volume, NULL);
 
 		gtk_widget_set_visible(priv->control_box, controls_visible);
 		gtk_widget_set_visible(priv->playlist, priv->playlist_visible);
-		gtk_window_resize(GTK_WINDOW(wnd), width, height);
+		gtk_window_set_default_size(GTK_WINDOW(wnd), width, height);
 		gtk_paned_set_position
-			(GTK_PANED(priv->vid_area_paned), handle_pos);
+			(GTK_PANED(priv->video_area_paned), handle_pos);
 
 		if(maximized)
 		{
@@ -607,7 +735,7 @@ celluloid_main_window_update_track_list(	CelluloidMainWindow *wnd,
 	priv->track_list = track_list;
 
 	celluloid_video_area_update_track_list
-		(CELLULOID_VIDEO_AREA(priv->vid_area), track_list);
+		(CELLULOID_VIDEO_AREA(priv->video_area), track_list);
 
 	if(celluloid_main_window_get_csd_enabled(wnd))
 	{
@@ -645,7 +773,7 @@ celluloid_main_window_update_disc_list(	CelluloidMainWindow *wnd,
 		celluloid_header_bar_update_disc_list
 			(CELLULOID_HEADER_BAR(priv->header_bar), disc_list);
 		celluloid_video_area_update_disc_list
-			(CELLULOID_VIDEO_AREA(priv->vid_area), disc_list);
+			(CELLULOID_VIDEO_AREA(priv->video_area), disc_list);
 	}
 	else
 	{
@@ -675,13 +803,13 @@ celluloid_main_window_resize_video_area(	CelluloidMainWindow *wnd,
 	 * fullscreen is unaffected, but it doesn't make sense to resize there
 	 * either.
 	 */
-	if(	!get_private(wnd)->fullscreen &&
+	if(	!gtk_window_is_fullscreen(GTK_WINDOW(wnd)) &&
 		!gtk_window_is_maximized(GTK_WINDOW(wnd)) )
 	{
 		CelluloidMainWindowPrivate *priv = get_private(wnd);
 
-		g_signal_connect(	priv->vid_area,
-					"size-allocate",
+		g_signal_connect(	priv->video_area,
+					"resize",
 					G_CALLBACK(resize_video_area_finalize),
 					wnd );
 
@@ -693,7 +821,7 @@ celluloid_main_window_resize_video_area(	CelluloidMainWindow *wnd,
 		 * resize_video_area_finalize() will be called so that the event handler
 		 * will be disconnected.
 		 */
-		gtk_widget_queue_allocate(priv->vid_area);
+		gtk_widget_queue_allocate(priv->video_area);
 	}
 }
 
@@ -720,37 +848,30 @@ celluloid_main_window_set_playlist_visible(	CelluloidMainWindow *wnd,
 {
 	CelluloidMainWindowPrivate *priv = get_private(wnd);
 
-	if(visible != priv->playlist_visible && !priv->fullscreen)
+	if(	visible != priv->playlist_visible &&
+		!gtk_window_is_fullscreen(GTK_WINDOW(wnd)))
 	{
-		GdkWindow *gdk_window;
-		GdkWindowState window_state;
 		gboolean resize;
 		gint handle_pos;
 		gint width;
 		gint height;
 
-		gdk_window = gtk_widget_get_window(GTK_WIDGET(wnd));
-		window_state = gdk_window_get_state(gdk_window);
-		resize = window_state &
-			(	GDK_WINDOW_STATE_TOP_RESIZABLE |
-				GDK_WINDOW_STATE_RIGHT_RESIZABLE |
-				GDK_WINDOW_STATE_BOTTOM_RESIZABLE |
-				GDK_WINDOW_STATE_LEFT_RESIZABLE );
+		resize = gtk_window_get_resizable(GTK_WINDOW(wnd));
 		handle_pos =	gtk_paned_get_position
-				(GTK_PANED(priv->vid_area_paned));
+				(GTK_PANED(priv->video_area_paned));
 
-		gtk_window_get_size(GTK_WINDOW(wnd), &width, &height);
+		gtk_window_get_default_size(GTK_WINDOW(wnd), &width, &height);
 
 		if(priv->playlist_first_toggle && visible)
 		{
-			gint new_pos = width-(resize?0:priv->playlist_width);
+			gint new_pos = width - (resize ? 0 : priv->playlist_width);
 
 			gtk_paned_set_position
-				(GTK_PANED(priv->vid_area_paned), new_pos);
+				(GTK_PANED(priv->video_area_paned), new_pos);
 		}
 		else if(!visible)
 		{
-			priv->playlist_width = width-handle_pos;
+			priv->playlist_width = width - handle_pos;
 		}
 
 		priv->playlist_visible = visible;
@@ -761,12 +882,11 @@ celluloid_main_window_set_playlist_visible(	CelluloidMainWindow *wnd,
 			gint new_width;
 
 			new_width =	visible?
-					width+priv->playlist_width:
+					width + priv->playlist_width:
 					handle_pos;
 
-			gtk_window_resize(	GTK_WINDOW(wnd),
-						new_width,
-						height );
+			gtk_window_set_default_size
+				(GTK_WINDOW(wnd), new_width, height);
 		}
 
 		priv->playlist_first_toggle = FALSE;
@@ -786,17 +906,27 @@ celluloid_main_window_set_controls_visible(	CelluloidMainWindow *wnd,
 	GSettings *settings = g_settings_new(CONFIG_WIN_STATE);
 	CelluloidMainWindowPrivate *priv = get_private(wnd);
 	const gboolean floating = priv->use_floating_controls;
+	const gboolean fullscreen = gtk_window_is_fullscreen(GTK_WINDOW(wnd));
 
 	gtk_widget_set_visible
-		(GTK_WIDGET(priv->control_box), visible && !floating);
-	g_settings_set_boolean
-		(settings, "show-controls", visible);
+		(	GTK_WIDGET(priv->control_box),
+			visible && !fullscreen && !floating );
+	celluloid_video_area_set_control_box_visible
+		(	CELLULOID_VIDEO_AREA(priv->video_area),
+			visible && (fullscreen || floating) );
 
+	g_settings_set_boolean (settings, "show-controls", visible);
 	g_clear_object(&settings);
 }
 
 gboolean
 celluloid_main_window_get_controls_visible(CelluloidMainWindow *wnd)
 {
-	return gtk_widget_get_visible(GTK_WIDGET(get_private(wnd)->control_box));
+	CelluloidMainWindowPrivate *priv = get_private(wnd);
+	CelluloidVideoArea *video_area = CELLULOID_VIDEO_AREA(priv->video_area);
+	const gboolean floating = priv->use_floating_controls;
+
+	return	(floating || gtk_window_is_fullscreen(GTK_WINDOW(wnd))) ?
+		celluloid_video_area_get_control_box_visible(video_area) :
+		gtk_widget_get_visible(GTK_WIDGET(priv->control_box));
 }
