@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2017-2020 gnome-mpv
+ * Copyright (c) 2017-2022 gnome-mpv
  *
  * This file is part of Celluloid.
  *
@@ -23,13 +23,47 @@
 
 #include "celluloid-option-parser.h"
 #include "celluloid-player.h"
-#include "celluloid-player-private.h"
 #include "celluloid-player-options.h"
 #include "celluloid-marshal.h"
 #include "celluloid-metadata-cache.h"
-#include "celluloid-mpv-wrapper.h"
+#include "celluloid-mpv.h"
 #include "celluloid-def.h"
 
+#define get_private(player) \
+	((CelluloidPlayerPrivate *)celluloid_player_get_instance_private(CELLULOID_PLAYER(player)))
+
+typedef struct _CelluloidPlayerPrivate CelluloidPlayerPrivate;
+
+enum
+{
+	PROP_0,
+	PROP_PLAYLIST,
+	PROP_METADATA,
+	PROP_CHAPTER_LIST,
+	PROP_TRACK_LIST,
+	PROP_DISC_LIST,
+	PROP_EXTRA_OPTIONS,
+	N_PROPERTIES
+};
+
+struct _CelluloidPlayerPrivate
+{
+	CelluloidMpv parent;
+	CelluloidMetadataCache *cache;
+	GVolumeMonitor *monitor;
+	GPtrArray *playlist;
+	GPtrArray *metadata;
+	GPtrArray *chapter_list;
+	GPtrArray *track_list;
+	GPtrArray *disc_list;
+	GHashTable *log_levels;
+	gboolean loaded;
+	gboolean new_file;
+	gboolean init_vo_config;
+	gchar *tmp_input_config;
+	gchar *extra_options;
+};
+
 static void
 set_property(	GObject *object,
 		guint property_id,
@@ -42,6 +76,9 @@ get_property(	GObject *object,
 		GValue *value,
 		GParamSpec *pspec );
 
+static void
+dispose(GObject *object);
+
 static void
 finalize(GObject *object);
 
@@ -96,6 +133,9 @@ load_input_config_file(CelluloidPlayer *player);
 static void
 load_scripts(CelluloidPlayer *player);
 
+static CelluloidChapter *
+convert_mpv_chapter(mpv_node_list *node);
+
 static CelluloidTrack *
 parse_track_entry(mpv_node_list *node);
 
@@ -114,6 +154,9 @@ update_playlist(CelluloidPlayer *player);
 static void
 update_metadata(CelluloidPlayer *player);
 
+static void
+update_chapter_list(CelluloidPlayer *player);
+
 static void
 update_track_list(CelluloidPlayer *player);
 
@@ -145,6 +188,7 @@ set_property(	GObject *object,
 	{
 		case PROP_PLAYLIST:
 		case PROP_METADATA:
+		case PROP_CHAPTER_LIST:
 		case PROP_TRACK_LIST:
 		case PROP_DISC_LIST:
 		g_critical("Attempted to set read-only property");
@@ -179,6 +223,10 @@ get_property(	GObject *object,
 		g_value_set_pointer(value, priv->metadata);
 		break;
 
+		case PROP_CHAPTER_LIST:
+		g_value_set_pointer(value, priv->chapter_list);
+		break;
+
 		case PROP_TRACK_LIST:
 		g_value_set_pointer(value, priv->track_list);
 		break;
@@ -197,6 +245,17 @@ get_property(	GObject *object,
 	}
 }
 
+static void
+dispose(GObject *object)
+{
+	CelluloidPlayerPrivate *priv = get_private(object);
+
+	g_clear_object(&priv->cache);
+	g_clear_object(&priv->monitor);
+
+	G_OBJECT_CLASS(celluloid_player_parent_class)->dispose(object);
+}
+
 static void
 finalize(GObject *object)
 {
@@ -211,6 +270,7 @@ finalize(GObject *object)
 	g_free(priv->extra_options);
 	g_ptr_array_free(priv->playlist, TRUE);
 	g_ptr_array_free(priv->metadata, TRUE);
+	g_ptr_array_free(priv->chapter_list, TRUE);
 	g_ptr_array_free(priv->track_list, TRUE);
 	g_ptr_array_free(priv->disc_list, TRUE);
 
@@ -388,6 +448,10 @@ mpv_property_changed(CelluloidMpv *mpv, const gchar *name, gpointer value)
 	{
 		update_metadata(player);
 	}
+	else if(g_strcmp0(name, "chapter-list") == 0)
+	{
+		update_chapter_list(player);
+	}
 	else if(g_strcmp0(name, "track-list") == 0)
 	{
 		update_track_list(player);
@@ -397,7 +461,6 @@ mpv_property_changed(CelluloidMpv *mpv, const gchar *name, gpointer value)
 		if(priv->init_vo_config)
 		{
 			priv->init_vo_config = FALSE;
-			load_scripts(player);
 			load_from_playlist(player);
 		}
 	}
@@ -412,13 +475,15 @@ observe_properties(CelluloidMpv *mpv)
 	celluloid_mpv_observe_property(mpv, 0, "aid", MPV_FORMAT_STRING);
 	celluloid_mpv_observe_property(mpv, 0, "vid", MPV_FORMAT_STRING);
 	celluloid_mpv_observe_property(mpv, 0, "sid", MPV_FORMAT_STRING);
+	celluloid_mpv_observe_property(mpv, 0, "chapter-list", MPV_FORMAT_NODE);
 	celluloid_mpv_observe_property(mpv, 0, "chapters", MPV_FORMAT_INT64);
 	celluloid_mpv_observe_property(mpv, 0, "core-idle", MPV_FORMAT_FLAG);
 	celluloid_mpv_observe_property(mpv, 0, "idle-active", MPV_FORMAT_FLAG);
 	celluloid_mpv_observe_property(mpv, 0, "border", MPV_FORMAT_FLAG);
 	celluloid_mpv_observe_property(mpv, 0, "fullscreen", MPV_FORMAT_FLAG);
 	celluloid_mpv_observe_property(mpv, 0, "pause", MPV_FORMAT_FLAG);
-	celluloid_mpv_observe_property(mpv, 0, "loop", MPV_FORMAT_STRING);
+	celluloid_mpv_observe_property(mpv, 0, "loop-file", MPV_FORMAT_STRING);
+	celluloid_mpv_observe_property(mpv, 0, "loop-playlist", MPV_FORMAT_STRING);
 	celluloid_mpv_observe_property(mpv, 0, "duration", MPV_FORMAT_DOUBLE);
 	celluloid_mpv_observe_property(mpv, 0, "media-title", MPV_FORMAT_STRING);
 	celluloid_mpv_observe_property(mpv, 0, "metadata", MPV_FORMAT_NODE);
@@ -467,7 +532,7 @@ apply_default_options(CelluloidMpv *mpv)
 			{"config-dir", config_dir},
 			{"watch-later-directory", watch_dir},
 			{"screenshot-directory", screenshot_dir},
-			{"screenshot-template", "celluloid-shot%n"},
+			{"screenshot-template", "%f-%P"},
 			{NULL, NULL} };
 
 	for(gint i = 0; options[i].name; i++)
@@ -488,8 +553,6 @@ static void
 initialize(CelluloidMpv *mpv)
 {
 	CelluloidPlayer *player = CELLULOID_PLAYER(mpv);
-	GSettings *win_settings = g_settings_new(CONFIG_WIN_STATE);
-	gdouble volume = g_settings_get_double(win_settings, "volume")*100;
 
 	apply_default_options(mpv);
 	load_config_file(mpv);
@@ -499,10 +562,7 @@ initialize(CelluloidMpv *mpv)
 
 	CELLULOID_MPV_CLASS(celluloid_player_parent_class)->initialize(mpv);
 
-	g_debug("Setting volume to %f", volume);
-	celluloid_mpv_set_property(mpv, "volume", MPV_FORMAT_DOUBLE, &volume);
-
-	g_object_unref(win_settings);
+	load_scripts(player);
 }
 
 static gint
@@ -517,15 +577,25 @@ apply_options_array_string(CelluloidMpv *mpv, const gchar *args)
 
 		args = parse_option(args, &key, &value);
 
-		g_debug("Applying option: --%s=%s", key, value);
-
-		if(celluloid_mpv_set_option_string(mpv, key, value) < 0)
+		if(key && *key)
 		{
-			fail_count++;
+			g_debug("Applying option: --%s=%s", key, value);
 
-			g_warning(	"Failed to apply option: --%s=%s\n",
-					key,
-					value );
+			if(celluloid_mpv_set_option_string(mpv, key, value) < 0)
+			{
+				fail_count++;
+
+				g_warning(	"Failed to apply option: --%s=%s\n",
+						key,
+						value );
+			}
+		}
+		else
+		{
+			g_warning("Failed to parse options");
+
+			args = NULL;
+			fail_count++;
 		}
 
 		g_free(key);
@@ -676,18 +746,22 @@ load_input_conf(CelluloidPlayer *player, const gchar *input_conf)
 	{
 		const gsize buf_size = 65536;
 		void *buf = g_malloc(buf_size);
-		FILE *src_file = g_fopen(input_conf, "r");
+		GFile *src_file = g_file_new_for_uri(input_conf);
+		GFileInputStream *src_stream = g_file_read(src_file, NULL, NULL);
 		gsize read_size = buf_size;
 
-		if(!src_file)
+		if(!src_stream)
 		{
 			g_warning(	"Cannot open input config file: %s",
 					input_conf );
 		}
 
-		while(src_file && read_size == buf_size)
+		while(src_stream && read_size == buf_size)
 		{
-			read_size = fread(buf, 1, buf_size, src_file);
+			read_size =
+				(gsize)
+				g_input_stream_read
+				(G_INPUT_STREAM(src_stream), buf, buf_size, NULL, NULL);
 
 			if(read_size != fwrite(buf, 1, read_size, tmp_file))
 			{
@@ -699,6 +773,8 @@ load_input_conf(CelluloidPlayer *player, const gchar *input_conf)
 
 		g_info("Loaded input config file: %s", input_conf);
 
+		g_clear_object(&src_stream);
+		g_object_unref(src_file);
 		g_free(buf);
 	}
 
@@ -715,9 +791,14 @@ load_config_file(CelluloidMpv *mpv)
 		gchar *mpv_conf =	g_settings_get_string
 					(settings, "mpv-config-file");
 
-		g_info("Loading config file: %s", mpv_conf);
-		celluloid_mpv_load_config_file(mpv, mpv_conf);
+		GFile *file = g_file_new_for_uri(mpv_conf);
+		gchar *path = g_file_get_path(file);
 
+		g_info("Loading config file: %s", path);
+		celluloid_mpv_load_config_file(mpv, path);
+
+		g_free(path);
+		g_object_unref(file);
 		g_free(mpv_conf);
 	}
 
@@ -785,6 +866,26 @@ load_scripts(CelluloidPlayer *player)
 	g_free(path);
 }
 
+static CelluloidChapter *
+convert_mpv_chapter(mpv_node_list *node)
+{
+	CelluloidChapter *chapter = celluloid_chapter_new();
+
+	for(gint i = 0; i < node->num; i++)
+	{
+		if(g_strcmp0(node->keys[i], "title") == 0)
+		{
+			chapter->title = g_strdup(node->values[i].u.string);
+		}
+		else if(g_strcmp0(node->keys[i], "time") == 0)
+		{
+			chapter->time = node->values[i].u.double_;
+		}
+	}
+
+	return chapter;
+}
+
 static CelluloidTrack *
 parse_track_entry(mpv_node_list *node)
 {
@@ -938,7 +1039,7 @@ update_metadata(CelluloidPlayer *player)
 
 	org_list = metadata.u.list;
 
-	if(metadata.format == MPV_FORMAT_NODE_MAP && org_list->num > 0)
+	if(metadata.format == MPV_FORMAT_NODE_MAP)
 	{
 		for(gint i = 0; i < org_list->num; i++)
 		{
@@ -971,6 +1072,38 @@ update_metadata(CelluloidPlayer *player)
 	}
 }
 
+static void
+update_chapter_list(CelluloidPlayer *player)
+{
+	CelluloidPlayerPrivate *priv = get_private(player);
+	mpv_node_list *org_list = NULL;
+	mpv_node chapter_list;
+
+	g_ptr_array_set_size(priv->chapter_list, 0);
+	celluloid_mpv_get_property(	CELLULOID_MPV(player),
+					"chapter-list",
+					MPV_FORMAT_NODE,
+					&chapter_list );
+
+	org_list = chapter_list.u.list;
+
+	if(chapter_list.format == MPV_FORMAT_NODE_ARRAY)
+	{
+		for(gint i = 0; i < org_list->num; i++)
+		{
+			CelluloidChapter *chapter =
+				convert_mpv_chapter
+				(org_list->values[i].u.list);
+
+			g_ptr_array_add(priv->chapter_list, chapter);
+		}
+
+		mpv_free_node_contents(&chapter_list);
+
+		g_object_notify(G_OBJECT(player), "chapter-list");
+	}
+}
+
 static void
 update_track_list(CelluloidPlayer *player)
 {
@@ -1022,7 +1155,7 @@ cache_update_handler(	CelluloidMetadataCache *cache,
 			entry->title = g_strdup(cache_entry->title);
 
 			g_signal_emit_by_name
-				(data, "metadata-update", (gint64)i);
+				(data, "metadata-cache-update", (gint64)i);
 		}
 	}
 }
@@ -1138,6 +1271,7 @@ celluloid_player_class_init(CelluloidPlayerClass *klass)
 	mpv_class->reset = reset;
 	obj_class->set_property = set_property;
 	obj_class->get_property = get_property;
+	obj_class->dispose = dispose;
 	obj_class->finalize = finalize;
 
 	pspec = g_param_spec_pointer
@@ -1154,6 +1288,13 @@ celluloid_player_class_init(CelluloidPlayerClass *klass)
 			G_PARAM_READABLE );
 	g_object_class_install_property(obj_class, PROP_METADATA, pspec);
 
+	pspec = g_param_spec_pointer
+		(	"chapter-list",
+			"Chapter list",
+			"List of chapters of the current file",
+			G_PARAM_READABLE );
+	g_object_class_install_property(obj_class, PROP_CHAPTER_LIST, pspec);
+
 	pspec = g_param_spec_pointer
 		(	"track-list",
 			"Track list",
@@ -1185,7 +1326,7 @@ celluloid_player_class_init(CelluloidPlayerClass *klass)
 			g_cclosure_marshal_VOID__VOID,
 			G_TYPE_NONE,
 			0 );
-	g_signal_new(	"metadata-update",
+	g_signal_new(	"metadata-cache-update",
 			G_TYPE_FROM_CLASS(klass),
 			G_SIGNAL_RUN_FIRST,
 			0,
@@ -1208,6 +1349,8 @@ celluloid_player_init(CelluloidPlayer *player)
 				((GDestroyNotify)celluloid_playlist_entry_free);
 	priv->metadata =	g_ptr_array_new_with_free_func
 				((GDestroyNotify)celluloid_metadata_entry_free);
+	priv->chapter_list =	g_ptr_array_new_with_free_func
+				((GDestroyNotify)celluloid_chapter_free);
 	priv->track_list =	g_ptr_array_new_with_free_func
 				((GDestroyNotify)celluloid_track_free);
 	priv->disc_list =	g_ptr_array_new_with_free_func
@@ -1281,23 +1424,15 @@ celluloid_player_remove_playlist_entry(CelluloidPlayer *player, gint64 position)
 	CelluloidMpv *mpv = CELLULOID_MPV(player);
 	gboolean idle_active =	celluloid_mpv_get_property_flag
 				(mpv, "idle-active");
-	gint64 playlist_count = 0;
 
-	celluloid_mpv_get_property
-		(mpv, "playlist-count", MPV_FORMAT_INT64, &playlist_count);
-
-	/* mpv doesn't send playlist change signal before going idle when the
-	 * last playlist entry is removed, so the internal playlist needs to be
-	 * directly updated.
-	 */
-	if(idle_active || playlist_count == 1)
+	if(idle_active)
 	{
-		g_ptr_array_remove_index(	get_private(player)->playlist,
-						(guint)position );
+		CelluloidPlayerPrivate *priv = get_private(player);
+
+		g_ptr_array_remove_index(priv->playlist, (guint)position);
 		g_object_notify(G_OBJECT(player), "playlist");
 	}
-
-	if(!idle_active)
+	else
 	{
 		const gchar *cmd[] = {"playlist_remove", NULL, NULL};
 		gchar *index_str =	g_strdup_printf
@@ -1321,18 +1456,17 @@ celluloid_player_move_playlist_entry(	CelluloidPlayer *player,
 
 	if(idle_active)
 	{
-		GPtrArray *playlist;
-		CelluloidPlaylistEntry **entry;
+		GPtrArray *playlist =
+			get_private(player)->playlist;
+		const gint64 entry_index =
+			(src > dst) ? src - 1 : src;
+		CelluloidPlaylistEntry *entry =
+			g_ptr_array_index(playlist, entry_index);
 
-		playlist = get_private(player)->playlist;
-		entry = (CelluloidPlaylistEntry **)
-			&g_ptr_array_index(playlist, src);
+		g_ptr_array_insert(playlist, (gint)dst, entry);
+		g_ptr_array_steal_index(playlist, (guint)src);
 
-		g_ptr_array_insert(playlist, (gint)dst, *entry);
-
-		/* Prevent the entry from being freed */
-		*entry = NULL;
-		g_ptr_array_remove_index(playlist, (guint)((src > dst)?--src:src));
+		g_object_notify(G_OBJECT(player), "playlist");
 	}
 	else
 	{
